# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Table manager.

This tool allows for managing table data as NumPy structured arrays.
It includes functions for field calculation, join and pivot tables.
Also, functions to manage tables used in other tools are included.
Tables can be exported to csv files.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # open a file
    >>> file1 = 'file1.csv'
    >>> table1 = rs.table_manager.open_file('file1.csv', field_names=['field1', 'field2'])
    >>> # perform a calculation
    >>> calculation = rs.table_manager.calculate(
    ... matrix=table1, expression_string='"field1" * 1.5',
    ... output_field_name='calc'
    ... )
    >>> # export the table to csv
    >>> rs.table_manager.export_table(
    ... matrix=calculation, output_path='output.csv',
    ... fields=['field1', 'calc'], separator=';', decimal_separator='.'
    ... )
"""  # noqa: E501

import io
import itertools
import os
from typing import Union, Optional

import numpy as np
from numpy.lib import recfunctions as rfn

from remotior_sensus.core import configurations as cfg
from remotior_sensus.util import files_directories

# test issues
try:
    if cfg.gdal_path is not None:
        os.add_dll_directory(cfg.gdal_path)
except Exception as error:
    cfg.logger.log.error(str(error))

try:
    from osgeo import ogr
except Exception as error:
    cfg.logger.log.error(str(error))

try:
    assert rfn.assign_fields_by_name
    assert rfn.require_fields
except Exception as error:
    cfg.logger.log.error(str(error))


def open_file(
        file_path: str, separators: Optional[Union[str, list]] = None,
        field_names: Optional[list] = None,
        skip_first_line: Optional[bool] = True,
        progress_message: Optional[bool] = True
) -> np.recarray:
    """Opens a file.

    Opens a file by reading the content and creating a table (which is a 
    NumPy structured array).
    File formats csv and dbf are supported.

    Args:
        file_path: path of output file.
        separators: list of characters used as separator in csv files; default is tab and comma.
        field_names: list of strings to be used as field names.
        skip_first_line: skip the first line in csv files, in case the first line contains field names.
        progress_message: if True then start progress message; 
            if False does not start the progress message (useful if launched from other tools).

    Returns:
        Table as NumPy structured array.

    Examples:
        Open a file
            >>> table_1 = open_file('file.csv')
    """  # noqa: E501
    if files_directories.file_extension(file_path) == cfg.dbf_suffix:
        table = _open_dbf(
            file_path=file_path, field_name_list=field_names,
            progress_message=progress_message
            )
    else:
        table = _open_csv(
            file_path=file_path, separators=separators,
            field_name_list=field_names, progress_message=progress_message,
            skip_first_line=skip_first_line
            )
    return table


# open dbf file
def _open_dbf(file_path, field_name_list=None, progress_message=True):
    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(process='open_dbf', message='starting', start=True)
    cfg.logger.log.debug('file_path: %s' % file_path)
    input_file = ogr.Open(file_path)
    i_layer = input_file.GetLayer()
    i_layer_def = i_layer.GetLayerDefn()
    field_count = i_layer_def.GetFieldCount()
    # fields
    dtype_list = []
    for c in range(field_count):
        tp = i_layer_def.GetFieldDefn(c).GetTypeName()
        width = i_layer_def.GetFieldDefn(c).GetWidth()
        if tp == 'Integer':
            dtype = 'int64'
        elif tp == 'Integer64':
            dtype = 'int64'
        elif tp == 'Real':
            dtype = 'float64'
        else:
            dtype = 'U%s' % width
        try:
            dtype_list.append((field_name_list[c], dtype))
        except Exception as err:
            str(err)
            dtype_list.append(
                (i_layer_def.GetFieldDefn(c).GetNameRef(), dtype)
            )
    mat_list = []
    i = 0
    feature_count = i_layer.GetFeatureCount()
    i_feature = i_layer.GetNextFeature()
    while i_feature:
        i += 1
        progress = int(100 * i / feature_count)
        if progress_message:
            cfg.progress.update(message='opening file', step=progress)
        else:
            cfg.progress.update(message='opening file', percentage=progress)
        row = []
        for c in range(field_count):
            f = i_feature.GetField(c)
            row.append(f)
        mat_list.append(row)
        i_feature = i_layer.GetNextFeature()
    array_ml = np.array(mat_list)
    rec_array = np.rec.fromarrays(array_ml.T, dtype=dtype_list)
    cfg.logger.log.info('end')
    if progress_message:
        cfg.progress.update(end=True)
    return rec_array


# open csv file
def _open_csv(
        file_path, separators=None, field_name_list=None,
        progress_message=True, skip_first_line=True
):
    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(process='open_csv', message='starting', start=True)
    if separators is None:
        separators = [cfg.tab_delimiter, cfg.comma_delimiter]
    cfg.logger.log.debug('file_path: %s' % file_path)
    with open(file_path, 'r') as file:
        # read file
        text = file.read()
        split_new_line = text.split(cfg.new_line)
        field_list = []
        data_list = []
        line_count = len(split_new_line)
        min_progress = 1
        max_progress = 80
        i = 0
        for line in split_new_line:
            i += 1
            if progress_message:
                cfg.progress.update(
                    message='opening file', step=i, steps=line_count,
                    minimum=min_progress,
                    maximum=max_progress, percentage=int(100 * i / line_count)
                )
            else:
                cfg.progress.update(
                    message='opening file',
                    percentage=int(100 * i / line_count)
                )
            if len(line) > 0:
                feature_list = [line]
                for s in separators:
                    features = []
                    for line_x in feature_list:
                        features.extend(line_x.split(s))
                    feature_list = features
                if i == 1:
                    if skip_first_line:
                        field_list = feature_list
                    else:
                        data_list.append(feature_list)
                        for f in range(1, len(feature_list) + 1):
                            field_list.append('field%s' % f)
                else:
                    data_list.append(feature_list)
    # create array
    array_data = np.array(data_list)
    dtype_list = []
    min_progress = 80
    max_progress = 99
    field_count = len(field_list)
    for t in range(len(field_list)):
        if progress_message:
            cfg.progress.update(
                message='processing data', step=t + 1, steps=field_count,
                minimum=min_progress, maximum=max_progress,
                percentage=int(100 * (t + 1) / line_count)
            )
        else:
            cfg.progress.update(
                message='processing data',
                percentage=int(100 * (t + 1) / line_count)
            )
        # check integer or float
        try:
            a = array_data[::, t].astype(np.float64)
            try:
                b = array_data[::, t].astype(np.int64)
                if np.nansum(a != b) == 0:
                    cfg.logger.log.debug('integer field: %s' % t)
                try:
                    dtype_list.append((field_name_list[t], 'int64'))
                except Exception as err:
                    str(err)
                    dtype_list.append((field_list[t], 'int64'))
            except Exception as err:
                str(err)
                cfg.logger.log.debug('float field: %s' % t)
                try:
                    dtype_list.append((field_name_list[t], 'float64'))
                except Exception as err:
                    str(err)
                    dtype_list.append((field_list[t], 'float64'))
        except Exception as err:
            str(err)
            cfg.logger.log.debug('string field: %s' % t)
            try:
                dtype_list.append((field_name_list[t], 'U64'))
            except Exception as err:
                str(err)
                dtype_list.append((field_list[t], 'U64'))
    rec_array = np.rec.fromarrays(array_data.T, dtype=dtype_list)
    if progress_message:
        cfg.progress.update(end=True)
    cfg.logger.log.info('end')
    return rec_array


# join matrices without duplicates
def join_matrices(
        matrix1, matrix2, field1_name, field2_name, join_type='leftouter',
        matrix1_postfix='_m1',
        matrix2_postfix='_m2', use_mask=False, progress_message=True
):
    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(
            process='join_matrices', message='starting', start=True
        )
    renamed_matrix1 = rfn.rename_fields(matrix1, {field1_name: field2_name})
    new_matrix = rfn.join_by(
        field2_name, renamed_matrix1, matrix2, jointype=join_type,
        r1postfix=matrix1_postfix,
        r2postfix=matrix2_postfix, asrecarray=True, usemask=use_mask
    )
    if progress_message:
        cfg.progress.update(end=True)
    cfg.logger.log.info('end')
    return new_matrix


# join tables
def join_tables(
        table1, table2, field1_name, field2_name, postfix='2',
        nodata_value=None, join_type='left', n_processes: int = None,
        progress_message=True, min_progress=None, max_progress=None
):
    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(
            process='join_tables', message='starting', start=True
        )
    if n_processes is None:
        n_processes = cfg.n_processes
    join_type = join_type.lower()
    if nodata_value is None:
        nodata_value = -9999
    cfg.multiprocess.join_tables_multiprocess(
        table1=table1, table2=table2, field1_name=field1_name,
        field2_name=field2_name, nodata_value=nodata_value,
        join_type=join_type, postfix=postfix, n_processes=n_processes,
        min_progress=min_progress, max_progress=max_progress
    )
    if progress_message:
        cfg.progress.update(end=True)
    cfg.logger.log.info('end')
    return cfg.multiprocess.output


# calculate pivot_60
def pivot_matrix(
        matrix, row_field, column_function_list, secondary_row_field_list=None,
        filter_string=None,
        nodata_value=-999, cross_matrix=None, progress_message=True
):
    """
    :param matrix: matrix array
    :param row_field: field out_column_name of values used as rows
    :param column_function_list: list of lists of column out_column_name and
        function on values and optional out dtype
    :param secondary_row_field_list: optional list of fields to be used as
        secondary rows
    :param filter_string: optional string for filtering input matrix values
    :param nodata_value: optional value for nodata
    :param cross_matrix: optional, if True the output table field names are
        only combination values
    :param progress_message: optional, if True display process message

    """

    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(
            process='pivot_matrix', message='starting', start=True
        )
    # matrix filter
    if filter_string is None:
        matrix_1 = matrix
    else:
        matrix_1 = eval('matrix[%s]' % filter_string)
    # rows
    row_values = np.unique(matrix_1[row_field])
    row_value_list: list = row_values.tolist()
    cfg.logger.log.debug(
        'column_function_list: %s; len(row_value_list): %s' % (
            str(column_function_list), len(row_value_list))
    )
    # primary rows
    if secondary_row_field_list is None:
        pivot = np.zeros([len(column_function_list) + 1, len(row_value_list)])
        secondary_row_list = None
    # secondary rows
    else:
        unique_value_list = []
        for field in secondary_row_field_list:
            unique_value_list.append(np.unique(matrix_1[field]).tolist())
        if len(secondary_row_field_list) == 1:
            secondary_row_list = unique_value_list[0]
        else:
            assert itertools
            secondary_row_list = eval(
                'list(itertools.product%s)' % str((tuple(unique_value_list)))
            )
        pivot = np.zeros(
            [len(column_function_list) * len(secondary_row_list) + 1,
             len(row_value_list)]
        )
        cfg.logger.log.debug(
            'secondary_row_list: %s' % str(secondary_row_list)
        )
    output_column_list = [(row_field, matrix_1[row_field].dtype)]
    # filters for secondary rows
    secondary_row_filter_list = []
    # primary rows
    if secondary_row_field_list is None:
        for column_function in column_function_list:
            out_column_name = '%s_%s' % (
                column_function[0], column_function[1])
            # output columns
            output_column_list.append(
                (out_column_name, matrix_1[column_function[0]].dtype)
            )
    # secondary rows
    else:
        for column_function in column_function_list:
            for combination in secondary_row_list:
                out_column_name = '%s_%s' % (
                    column_function[0], column_function[1].replace('.', ''))
                function_string = ['[ ']
                for n in range(len(secondary_row_field_list)):
                    try:
                        try:
                            float(combination[n])
                            comb_string = str(combination[n])
                        except Exception as err:
                            str(err)
                            comb_string = '"%s"' % combination[n]
                        if cross_matrix:
                            out_column_name = str(combination[n])
                        else:
                            out_column_name = '%s_%s%s' % (
                                out_column_name, secondary_row_field_list[n],
                                combination[n])
                        function_string.append(
                            '(r["%s"] == %s)' % (
                                secondary_row_field_list[n], comb_string)
                        )
                        function_string.append(' &')
                    except Exception as err:
                        str(err)
                        try:
                            float(combination)
                            comb_string = str(combination)
                        except Exception as err:
                            str(err)
                            comb_string = '"%s"' % combination
                        if cross_matrix:
                            out_column_name = str(combination)
                        else:
                            out_column_name = '%s_%s%s' % (
                                out_column_name, secondary_row_field_list[n],
                                combination)
                        function_string.append(
                            '(r["%s"] == %s)' % (
                                secondary_row_field_list[n], comb_string)
                        )
                        function_string.append(' &')
                if function_string[-1] == ' &':
                    function_string.pop(-1)
                function_string.append(']')
                secondary_row_filter_list.append(''.join(function_string))
                # output columns
                try:
                    output_column_list.append(
                        (out_column_name, column_function[2])
                    )
                except Exception as err:
                    str(err)
                    output_column_list.append(
                        (out_column_name, matrix_1[column_function[0]].dtype)
                    )
    cfg.logger.log.debug(
        'secondary_row_filter_list: %s; output_column_list: %s'
        % (str(secondary_row_filter_list), str(output_column_list))
    )
    pivot = np.rec.fromarrays(pivot, dtype=output_column_list)
    pivot[row_field] = row_values
    # populate table
    row_value_count = len(row_value_list)
    for v in range(row_value_count):
        if progress_message:
            cfg.progress.update(
                message='processing data',
                step=int(100 * (v + 1) / row_value_count),
                percentage=int(100 * (v + 1) / row_value_count)
            )
        else:
            cfg.progress.update(
                message='processing data',
                percentage=int(100 * (v + 1) / row_value_count)
            )
        r = matrix_1[matrix_1[row_field] == row_value_list[v]]
        assert r.shape
        d = 0
        for column in range(len(column_function_list)):
            operator = replace_numpy_operators(column_function_list[column][1])
            try:
                datatype = column_function_list[column][2]
            except Exception as err:
                str(err)
                datatype = matrix_1[column_function_list[column][0]].dtype
            if secondary_row_field_list is None:
                try:
                    s = eval(
                        '%s(r["%s"].astype("%s"))' % (
                            operator, column_function_list[column][0],
                            datatype)
                    )
                except Exception as err:
                    str(err)
                    s = nodata_value
                pivot[output_column_list[column + 1][0]][v] = s
            else:
                for _ in secondary_row_list:
                    try:
                        s = eval(
                            '%s((r["%s"]%s).astype("%s"))'
                            % (operator, column_function_list[column][0],
                               secondary_row_filter_list[d], datatype)
                        )
                    except Exception as err:
                        str(err)
                        s = nodata_value
                    pivot[output_column_list[d + 1][0]][v] = s
                    d += 1
    if progress_message:
        cfg.progress.update(end=True)
    cfg.logger.log.info('end')
    return pivot


# get values from matrix
def get_values(
        matrix, value_field, conditional_string=None, progress_message=True
):
    """
    :param matrix: input matrix
    :param value_field: string of value field name
    :param conditional_string: optional string used for condition where
        field must be referred to using 'field.' such as field.field_name
        or 'matrix.' such as matrix.field_name
    :param progress_message: optional, if True display process message
    """
    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(
            process='get_values', message='starting', start=True
        )
    # alias for fields
    field = matrix
    matrix_values = field[value_field]
    if conditional_string is None:
        values = matrix_values
    else:
        try:
            a = eval(conditional_string)
            values = matrix_values[a]
        except Exception as err:
            cfg.logger.log.error(str(err))
            return
    if progress_message:
        cfg.progress.update(end=True)
    cfg.logger.log.info('end')
    return values


# replace variables in expression for calculation
def replace_variables(matrix, expression_string):
    for field in columns(matrix):
        expression_string = expression_string.replace(
            '%s%s%s' % (
                cfg.variable_band_quotes, field, cfg.variable_band_quotes),
            'field[%s%s%s]' % (
                cfg.variable_band_quotes, field, cfg.variable_band_quotes)
        )
    cfg.logger.log.debug('expression_string: %s' % expression_string)
    return expression_string


# calculate single expression
def calculate(
        matrix, expression_string, output_field_name, progress_message=True
):
    """
    :param matrix: input matrix
    :param expression_string: string used for calculation where fields must
        be named using double quotes
        e.g."field_name" or using 'field.' such as field.field_name or
        'matrix.' such as matrix.field_name
    :param output_field_name: string of output field name
    :param progress_message: optional, if True display process message
    """
    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(
            process='calculate', message='starting', start=True
        )
    # alias for field in expression
    field = matrix
    expression_string = replace_variables(matrix, expression_string)
    # expose numpy functions
    log = np.log
    _log = log
    log10 = np.log10
    _log10 = log10
    sqrt = np.sqrt
    _sqrt = sqrt
    cos = np.cos
    _cos = cos
    arccos = np.arccos
    _arccos = arccos
    sin = np.sin
    _sin = sin
    arcsin = np.arcsin
    _arcsin = arcsin
    tan = np.tan
    _tan = tan
    arctan = np.arctan
    _arctan = arctan
    exp = np.exp
    _exp = exp
    min = np.nanmin
    _min = min
    max = np.nanmax
    _max = max
    sum = np.nansum
    _sum = sum
    percentile = np.nanpercentile
    _percentile = percentile
    median = np.nanmedian
    _median = median
    mean = np.nanmean
    _mean = mean
    std = np.nanstd
    _std = std
    where = np.where
    _where = where
    nan = np.nan
    _nan = nan
    matrix1 = None
    try:
        a = eval(expression_string)
        try:
            # existing field
            assert field[output_field_name]
            matrix1 = np.rec.array(np.copy(matrix))
            matrix1[output_field_name] = a
        except Exception as err:
            str(err)
            # new field
            matrix1 = append_field(matrix, output_field_name, a, a.dtype)
    except Exception as err:
        cfg.logger.log.error(str(err))
    if progress_message:
        cfg.progress.update(end=True)
    cfg.logger.log.info('end')
    return matrix1


# calculate multiple expressions
def calculate_multi(
        matrix, expression_string_list, output_field_name_list,
        progress_message=True
):
    """
    :param matrix: input matrix
    :param expression_string_list: list of strings used for calculation
        where fields must be named using double quotes
        e.g."field_name" or using 'field.' such as field.field_name or
        'matrix.' such as matrix.field_name
    :param output_field_name_list: list of strings of output field names
    :param progress_message: optional, if True display process message
    """
    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(
            process='calculate_multi', message='starting', start=True
        )
    # alias for field in expression
    field = matrix
    matrix1 = np.rec.array(np.copy(field))
    # expose numpy functions
    log = np.log
    _log = log
    log10 = np.log10
    _log10 = log10
    sqrt = np.sqrt
    _sqrt = sqrt
    cos = np.cos
    _cos = cos
    arccos = np.arccos
    _arccos = arccos
    sin = np.sin
    _sin = sin
    arcsin = np.arcsin
    _arcsin = arcsin
    tan = np.tan
    _tan = tan
    arctan = np.arctan
    _arctan = arctan
    exp = np.exp
    _exp = exp
    min = np.nanmin
    _min = min
    max = np.nanmax
    _max = max
    sum = np.nansum
    _sum = sum
    percentile = np.nanpercentile
    _percentile = percentile
    median = np.nanmedian
    _median = median
    mean = np.nanmean
    _mean = mean
    std = np.nanstd
    _std = std
    where = np.where
    _where = where
    nan = np.nan
    _nan = nan
    expression_count = len(expression_string_list)
    for e in range(expression_count):
        if progress_message:
            cfg.progress.update(
                message='processing data',
                step=int(100 * (e + 1) / expression_count),
                percentage=int(100 * (e + 1) / expression_count)
            )
        else:
            cfg.progress.update(
                message='processing data', percentage=int(
                    100 * (e + 1) / expression_count
                )
            )
        expression = replace_variables(matrix, expression_string_list[e])
        try:
            a = eval(expression)
            try:
                matrix1[output_field_name_list[e]] = a
            except Exception as err:
                str(err)
                matrix1 = append_field(
                    matrix1, output_field_name_list[e], a, a.dtype
                )
        except Exception as err:
            cfg.logger.log.error(str(err))
    if progress_message:
        cfg.progress.update(end=True)
    cfg.logger.log.info('end')
    return matrix1


# create new matrix selecting columns by name and optionally rename output
# columns
def redefine_matrix_columns(
        matrix, input_field_names, output_field_names=None,
        progress_message=True
):
    """
    :param matrix: matrix array
    :param input_field_names: list of field names to be included in output
    :param output_field_names: optional list of output field names with the
        same length as input_field_names
    :param progress_message: optional, if True display process message

    """

    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(
            process='redefine_matrix_columns', message='starting', start=True
        )
    field_list = []
    c = 0
    for field in input_field_names:
        if field in matrix.dtype.names:
            data_type = matrix[field].dtype
            field_list.append((field, data_type))
            c += 1
        else:
            cfg.logger.log.error('field %s not found' % field)
    matrix_f = define_fields(matrix, field_list)
    if output_field_names is not None:
        for f in range(len(output_field_names)):
            try:
                matrix_f = rename_field(
                    matrix_f, input_field_names[f], output_field_names[f]
                )
            except Exception as err:
                cfg.logger.log.error(str(err))
    if progress_message:
        cfg.progress.update(end=True)
    return matrix_f


# export matrix to csv, with field name list, and field decimals as integer
# or list
def matrix_to_csv(
        matrix, output_path, fields=None, field_decimals: Union[int, list] = 2,
        separator=None, decimal_separator=None, nodata_value=None,
        nodata_value_output='nan', progress_message=True
):
    """
    :param matrix: input matrix
    :param output_path: output path
    :param fields: optional list of fields to export
    :param field_decimals: integer number of decimals for decimal fields or
        list of integer values for decimal fields
    :param separator: separator of fields
    :param decimal_separator: optional decimal separator for float values
        replacing . character
    :param nodata_value: optional nodata value to be replaced
    :param nodata_value_output: optional string to replace nodata value in
        output
    :param progress_message:

    """

    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(
            process='matrix_to_csv', message='starting', start=True
        )
    if fields is None:
        fields = matrix.dtype.names
    if separator is None:
        separator = cfg.tab_delimiter
    dtypes = []
    for name in matrix.dtype.names:
        data_type = matrix[name].dtype
        if 'int' in str(data_type).lower() or '<i' in str(data_type).lower():
            data_type = np.dtype('int64')
        dtypes.append((name, str(data_type)))
    cfg.logger.log.debug('dtypes: %s' % str(dtypes))
    matrix_c = np.rec.array(np.copy(matrix))
    cfg.logger.log.debug('matrix_c.shape: %s' % str(matrix_c.shape))
    matrix_c = define_fields(matrix_c, dtypes)
    # list of field formats
    format_list = []
    # list of fields
    field_list = []
    c = 0
    header = []
    # iterate fields
    for field in fields:
        cfg.logger.log.debug('field: %s' % str(field))
        # header
        header.append(field)
        header.append(separator)
        if field in matrix.dtype.names:
            data_type = matrix[field].dtype
            field_list.append((field, data_type))
            # integer fields
            if ('int' in str(data_type).lower()
                    or '<i' in str(data_type).lower()):
                matrix_c[field][
                    matrix[field] == nodata_value] = cfg.nodata_val_Int64
                format_list.append('%s')
            # float fields
            elif (str(data_type)[0].lower() == 'f'
                  or '<f' in str(data_type)[0].lower()
                  or ('>f' in str(data_type)[0].lower())):
                matrix_c[field][
                    matrix[field] == nodata_value] = cfg.nodata_val_Int64
                try:
                    format_list.append('%1.{}f'.format(field_decimals[c]))
                except Exception as err:
                    str(err)
                    try:
                        format_list.append('%1.{}f'.format(field_decimals))
                    except Exception as err:
                        str(err)
                        format_list.append('%1.2f')
            # string fields
            else:
                format_list.append('%s')
                matrix_c[field][matrix[field] == str(
                    nodata_value)] = str(cfg.nodata_val_Int64)
        else:
            cfg.logger.log.error('field %s not found' % field)
        c += 1
    header.pop(-1)
    header.append(cfg.new_line)
    joined_header = ''.join(header)
    matrix_format = define_fields(matrix_c, field_list)
    # export numpy matrix to text
    stream = io.StringIO()
    np.savetxt(stream, matrix_format, delimiter=separator, fmt=format_list)
    csv = '{}{}'.format(joined_header, stream.getvalue())
    if decimal_separator is not None:
        csv = csv.replace('.', decimal_separator)
    # replace nodata values
    if nodata_value_output is not None:
        if decimal_separator is None:
            s = '.'
        else:
            s = decimal_separator
        if type(field_decimals) is list:
            for fd in field_decimals[c]:
                csv = csv.replace(
                    '%s%s%s' % (cfg.nodata_val_Int64, s, '0' * fd),
                    str(nodata_value_output)
                )
        else:
            csv = csv.replace(
                '%s%s%s' % (cfg.nodata_val_Int64, s, '0' * field_decimals),
                str(nodata_value_output)
            )
        csv = csv.replace(str(cfg.nodata_val_Int64), str(nodata_value_output))
    files_directories.create_parent_directory(output_path)
    with open(output_path, 'w') as file:
        file.write(csv)
    if progress_message:
        cfg.progress.update(end=True)
    cfg.logger.log.info('end')


"""Alias for matrix_to_csv."""
export_table = matrix_to_csv


# replace operators
def replace_numpy_operators(expression):
    if expression == 'sum':
        f = 'np.nansum'
    elif expression == 'min':
        f = 'np.nanmin'
    elif expression == 'max':
        f = 'np.nanmax'
    elif expression == 'percentile':
        f = 'np.nanpercentile'
    elif expression == 'median':
        f = 'np.nanmedian'
    elif expression == 'mean':
        f = 'np.nanmean'
    elif expression == 'std':
        f = 'np.nanstd'
    elif expression == 'log':
        f = 'np.log'
    elif expression == 'log10':
        f = 'np.log10'
    elif expression == 'sqrt':
        f = 'np.sqrt'
    elif expression == 'cos':
        f = 'np.cos'
    elif expression == 'sin':
        f = 'np.sin'
    elif expression == 'tan':
        f = 'np.tan'
    elif expression == 'exp':
        f = 'np.exp'
    elif expression == 'where':
        f = 'np.where'
    else:
        f = expression
    return f


# list column names
def columns(matrix):
    cfg.logger.log.debug('matrix.dtype.names: %s' % str(matrix.dtype.names))
    return matrix.dtype.names


# rename field matrices
def rename_field(matrix, old_field, new_field):
    new_matrix = rfn.rename_fields(matrix, {old_field: new_field})
    cfg.logger.log.debug(
        'old_field: %s; new_field: %s' % (old_field, new_field)
    )
    return new_matrix


# append field matrix
def append_field(matrix, field_name, data, data_type):
    new_matrix = rfn.append_fields(
        matrix, field_name, data, data_type, usemask=False, asrecarray=True
    )
    cfg.logger.log.debug(
        'field_name: %s; data_type: %s' % (field_name, data_type)
    )
    return new_matrix


# define fields matrix by list of tuples (field, dataType)
def define_fields(matrix, field_list):
    matrix_with_fields = rfn.require_fields(matrix, field_list)
    new_matrix = np.rec.array(matrix_with_fields)
    cfg.logger.log.debug('field_list: %s' % field_list)
    return new_matrix


# append values to table
def append_values_to_table(matrix, value_list):
    array_1 = np.array(value_list)
    rec_array = np.rec.fromarrays(array_1, dtype=matrix.dtype)
    a = append_tables(matrix, rec_array)
    cfg.logger.log.debug('a.shape: %s' % a.shape)
    return a


# append two tables
def append_tables(matrix1, matrix2):
    a = rfn.stack_arrays((matrix1, matrix2), asrecarray=True, usemask=False)
    cfg.logger.log.debug('a.shape: %s' % a.shape)
    return a


# sort table
def sort_table_by_field(matrix, field_name):
    a = np.rec.array(np.copy(matrix))
    a.sort(order=field_name)
    cfg.logger.log.debug('a.shape: %s' % a.shape)
    return a


# add product to preprocess table
def add_product_to_preprocess(
        product_list, spacecraft_list, processing_level, band_name_list,
        product_path_list,
        scale_list, offset_list, nodata_list, date_list, k1_list, k2_list,
        band_number_list, e_sun_list, sun_elevation_list,
        earth_sun_distance_list
):
    dtype_list = [
        ('product', 'U64'), ('spacecraft', 'U64'), ('processing_level', 'U64'),
        ('band_name', 'U128'), ('product_path', 'U1024'), ('scale', 'float64'),
        ('offset', 'float64'), ('nodata', 'float64'),
        ('date', 'datetime64[D]'), ('k1', 'float64'), ('k2', 'float64'),
        ('band_number', 'U64'), ('e_sun', 'float64'),
        ('sun_elevation', 'float64'), ('earth_sun_distance', 'float64')
    ]
    rec_array = np.rec.fromarrays(
        np.array(
            [product_list, spacecraft_list, processing_level, band_name_list,
             product_path_list, scale_list,
             offset_list, nodata_list, date_list, k1_list, k2_list,
             band_number_list, e_sun_list, sun_elevation_list,
             earth_sun_distance_list]
        ), dtype=dtype_list
    )
    cfg.logger.log.debug('rec_array.shape: %s' % rec_array.shape)
    return rec_array


# create product table
def create_product_table(
        product=None, product_id=None, acquisition_date=None, cloud_cover=None,
        zone_path=None,
        row=None, min_lat=None, min_lon=None, max_lat=None, max_lon=None,
        collection=None, size=None,
        preview=None, uid=None, image=None
):
    dtype_list = [
        ('product', 'U512'), ('image', 'U1024'), ('product_id', 'U512'),
        ('acquisition_date', 'datetime64[D]'), ('cloud_cover', 'int8'),
        ('zone_path', 'U8'), ('row', 'U8'), ('min_lat', 'float64'),
        ('min_lon', 'float64'), ('max_lat', 'float64'), ('max_lon', 'float64'),
        ('collection', 'U1024'), ('size', 'U512'), ('preview', 'U1024'),
        ('uid', 'U1024')
    ]
    rec_array = np.rec.fromrecords(
        [(
            product, image, product_id, acquisition_date, cloud_cover,
            zone_path, row, min_lat, min_lon, max_lat, max_lon,
            collection, size, preview, uid)], dtype=dtype_list
    )
    cfg.logger.log.debug('rec_array.shape: %s' % rec_array.shape)
    return rec_array


# create table of products
def stack_product_table(product_list):
    if product_list is None or len(product_list) == 0:
        return None
    table = rfn.stack_arrays(product_list, asrecarray=True, usemask=False)
    table.sort(order='product')
    cfg.logger.log.debug('table.shape: %s' % table.shape)
    return table


# create bandset table
def create_bandset_table(band_list):
    if band_list is None:
        return None
    table = rfn.stack_arrays(band_list, asrecarray=True, usemask=False)
    table.sort(order='band_number')
    cfg.logger.log.debug('table.shape: %s' % table.shape)
    return table


# add spectral signature to catalog table
def add_spectral_signature_to_catalog_table(
        signature_id=None, macroclass_id=0, class_id=0, class_name=None,
        previous_catalog=None, selected=1, min_dist_thr=0, max_like_thr=0,
        spec_angle_thr=0
):
    dtype_list = [('signature_id', 'U64'), ('macroclass_id', 'int16'),
                  ('class_id', 'int16'), ('class_name', 'U512'),
                  ('selected', 'byte'), ('min_dist_thr', 'float64'),
                  ('max_like_thr', 'float64'), ('spec_angle_thr', 'float64')]
    rec_array = np.rec.fromrecords(
        [(signature_id, int(macroclass_id), int(class_id), str(class_name),
          selected,
          min_dist_thr, max_like_thr, spec_angle_thr)],
        dtype=dtype_list
    )
    # add to previous bandset catalog table
    if previous_catalog is not None:
        # replace signature_id
        catalog = previous_catalog[
            previous_catalog['signature_id'] != signature_id]
        table = rfn.stack_arrays(
            [catalog, rec_array], asrecarray=True, usemask=False
        )
    else:
        table = rec_array
    return table


# create spectral signature table
def create_spectral_signature_table(
        value_list, wavelength_list, standard_deviation_list=None
):
    dtype_list = [('value', 'float64'), ('wavelength', 'float64'),
                  ('standard_deviation', 'float64')]
    if standard_deviation_list is None:
        standard_deviation_list = [0] * len(value_list)
    data_list = [value_list, wavelength_list, standard_deviation_list]
    rec_array = np.rec.fromarrays(np.array(data_list), dtype=dtype_list)
    return rec_array


# create band table
def create_band_table(
        band_number=0, raster_band=None, path=None, absolute_path=None,
        name=None, wavelength=0, wavelength_unit=None, additive_factor=0,
        multiplicative_factor=1, date='1900-01-01', x_size=0,
        y_size=0, top=0, left=0, bottom=0, right=0, x_count=0, y_count=0,
        nodata=None, data_type=None, crs=None, root_directory=None,
        number_of_bands=0, x_block_size=0, y_block_size=0, scale=None,
        offset=None
):
    dtype_list = [
        ('band_number', 'int16'), ('raster_band', 'int16'), ('path', 'U1024'),
        ('absolute_path', 'U1024'), ('name', 'U256'),
        ('wavelength', 'float64'), ('wavelength_unit', 'U32'),
        ('additive_factor', 'float32'), ('multiplicative_factor', 'float32'),
        ('date', 'datetime64[D]'), ('x_size', 'float32'),
        ('y_size', 'float32'), ('top', 'float32'), ('left', 'float32'),
        ('bottom', 'float32'), ('right', 'float32'), ('x_count', 'int64'),
        ('y_count', 'int64'), ('nodata', 'int64'), ('data_type', 'U16'),
        ('crs', 'U1024'), ('root_directory', 'U1024'),
        ('number_of_bands', 'int16'), ('x_block_size', 'int64'),
        ('y_block_size', 'int64'), ('scale', 'int64'), ('offset', 'int64')
    ]
    if nodata is None:
        nodata = cfg.nodata_val_Int64
    else:
        try:
            _array = np.array([nodata], dtype=int)
        except Exception as err:
            cfg.logger.log.error('%s; nodata: %s' % (err, nodata))
            nodata = cfg.nodata_val_Int64
    if raster_band is None:
        raster_band = 1
    if scale is None:
        scale = 1
    if offset is None:
        offset = 0
    try:
        rec_array = np.rec.fromrecords(
            [(band_number, raster_band, path, absolute_path, name,
              wavelength, wavelength_unit, additive_factor,
              multiplicative_factor, date, x_size, y_size, top, left, bottom,
              right, x_count, y_count, nodata, data_type, crs, root_directory,
              number_of_bands, x_block_size, y_block_size, scale, offset)],
            dtype=dtype_list
        )
        cfg.logger.log.debug('rec_array.shape: %s' % rec_array.shape)
    except Exception as err:
        cfg.logger.log.error(str(err))
        rec_array = None
    return rec_array


# create bandset catalog table
def create_bandset_catalog_table(
        bandset_number=0, root_directory=None, date='NaT', bandset_uid=0,
        bandset_name=None, previous_catalog=None, crs=None,
        box_coordinate_left=None, box_coordinate_top=None,
        box_coordinate_right=None, box_coordinate_bottom=None
):
    dtype_list = [
        ('bandset_number', 'int64'), ('bandset_name', 'U512'),
        ('date', 'datetime64[D]'), ('root_directory', 'U1024'),
        ('crs', 'U1024'), ('box_coordinate_left', 'float64'),
        ('box_coordinate_top', 'float64'), ('box_coordinate_right', 'float64'),
        ('box_coordinate_bottom', 'float64'), ('uid', 'U64')
    ]
    rec_array = np.rec.fromrecords(
        [(bandset_number, bandset_name, date, root_directory, crs,
          box_coordinate_left, box_coordinate_top, box_coordinate_right,
          box_coordinate_bottom, bandset_uid)], dtype=dtype_list
    )
    # add to previous bandset catalog table
    if previous_catalog is not None:
        table = rfn.stack_arrays(
            [previous_catalog, rec_array], asrecarray=True, usemask=False
        )
        table.sort(order='bandset_number')
    else:
        table = rec_array
    return table


# find nearest value
def find_nearest_value(array, field_name, value, threshold=None):
    difference = np.abs(array[field_name] - value)
    if threshold is None:
        index = np.argmin(difference)
        output = array[index]
    else:
        if np.nanmin(difference) < threshold:
            index = np.argmin(difference)
            output = array[index]
        else:
            output = None
    return output



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.napoleon_type_aliases
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""BandSet Catalog manager.

Core class that manages a catalog of BandSets.
A :func:`~remotior_sensus.core.bandset.BandSet` is an object that includes
information about single bands.

The BandSet Catalog allows for the definition of multimple BandSets
by a reference number, and the access to BandSet functions.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # set lists of files
    >>> file_list = ['file_1.tif', 'file_2.tif', 'file_3.tif']
    >>> # create BandSet Catalog
    >>> catalog = rs.bandset_catalog()
    >>> catalog.create_bandset(paths=file_list)
    >>> # for instance get BandSet count
    >>> print(catalog.get_bandset_count())
    1
"""
from copy import deepcopy
from typing import Union, Optional

import numpy

from remotior_sensus.core import (
    configurations as cfg, messages, table_manager as tm
)
from remotior_sensus.core.bandset import BandSet


class BandSetCatalog(object):
    """Manages BandSets.

    This class manages BandSets through a catalog, defining attributes and
    properties of BandSets.

    Attributes:
        bandsets: dictionary of the actual BandSets
        bandsets_table: table of BandSets containing the properties thereof
        get: alias for get_bandset

    Examples:
        Create a BandSet Catalog
            >>> import remotior_sensus
            >>> rs = remotior_sensus.Session()
            >>> catalog = rs.bandset_catalog()
    """  # noqa: E501

    def __init__(self):
        """Initializes the catalog with an empty BandSet"""
        self.bandsets = {}
        self.bandsets_table = None
        self.current_bandset = 1
        # alias for get_bandset
        self.get = self.get_bandset
        # empty BandSet
        self._empty_bandset()

    def get_bandset_count(self) -> int:
        """Gets count of BandSets in the catalog.

        This function gets the count of BandSets present in the catalog.

        Returns:
            The integer number of BandSets.

        Examples:
            Count of BandSets present.
                >>> catalog = BandSetCatalog()
                >>> count = catalog.get_bandset_count()
                >>> print(count)
                1
        """
        return len(self.bandsets)

    @property
    def current_bandset(self) -> int:
        """Property that defines the current BandSet in the catalog.

        This property identifies a BandSet number which is considered
        current (i.e. active) in several other tools when
        no BandSet is specified.

        Returns:
            The integer number of current BandSet.

        Examples:
            Get current BandSet.
                >>> catalog = BandSetCatalog()
                >>> bandset_number = catalog.current_bandset
                >>> print(bandset_number)
                1

            Set current BandSet.
                >>> catalog = BandSetCatalog()
                >>> print(catalog.current_bandset)
                1
                >>> catalog.current_bandset = 2
                >>> print(catalog.current_bandset)
                2
         """
        return self._current_bandset

    @current_bandset.setter
    def current_bandset(self, bandset_number: int):
        if bandset_number > self.get_bandset_count():
            bandset_number = self.get_bandset_count()
        if bandset_number == 0:
            bandset_number = 1
        self._current_bandset = bandset_number

    def get_bandset_by_number(self, number: int) -> Union[None, BandSet]:
        """Get BandSet by number thereof.

        This function gets the BandSet by the number thereof.

        Args:
            number: number of the BandSet

        Returns:
            The BandSet identified by the number.

        Examples:
            Get the first BandSet.
                >>> catalog = BandSetCatalog()
                >>> bandset_1 = catalog.get_bandset_by_number(1)
                >>> print(bandset_1)
                BandSet object
        """
        cfg.logger.log.debug('number: %s' % number)
        bandset = self.get_bandset(number)
        return bandset

    def get_bandset_by_name(
            self, bandset_name: str, output_number: Optional[bool] = False
    ) -> Union[None, int, BandSet]:
        """Get BandSet by name thereof.

        This function gets the BandSet by the name thereof.

        Args:
            bandset_name: name of the BandSet
            output_number: if True then the output is the BandSet number,
                if False then the output is the BandSet

        Returns:
            The BandSet identified by the name.

        Examples:
            Get the number of the BandSet having the name 'example'.
                >>> catalog = BandSetCatalog()
                >>> bandset_number = catalog.get_bandset_by_name(
                bandset_name='example', output_number=True)
                >>> print(bandset_number)
                3

            Get the BandSet having the name 'example'.
                >>> catalog = BandSetCatalog()
                >>> bandset = catalog.get_bandset_by_name(
                bandset_name='example', output_number=False)
                >>> print(bandset)
                BandSet object
        """
        cfg.logger.log.debug(
            'bandset_name: %s; output_number: %s' % (
                bandset_name, output_number)
        )
        uid = self.bandsets_table[
            self.bandsets_table['bandset_name'] == bandset_name]['uid']
        if len(uid) == 0:
            return None
        else:
            if output_number:
                bandset_array = self.bandsets_table['bandset_number'][
                    self.bandsets_table['uid'] == str(uid[0])]
                if len(bandset_array) > 0:
                    return bandset_array[0]
            else:
                return self.bandsets[str(uid[0])]

    def get_bandset_catalog_attributes(
            self, bandset_number: int,
            attribute: Optional[str] = None
    ) -> Union[None, str, list]:
        """Get BandSet Catalog attributes by number of the BandSet.

        This function gets the BandSet Catalog attributes by number of the
            BandSet.

        Args:
            bandset_number: number of the BandSet
            attribute: string of the attribute name, if None then the list
                of attributes is returned

        Returns:
            The BandSet attribute or the list of attributes
                ('bandset_number', 'bandset_name', 'date', 'root_directory',
                'uid').

        Examples:
            Get the 'date' attribute of the BandSet 1.
                >>> catalog = BandSetCatalog()
                >>> date = catalog.get_bandset_catalog_attributes(
                bandset_number=1, attribute='date')
                >>> print(date)
                2000-12-31

            Get the list of attributes of the BandSet 1.
                >>> catalog = BandSetCatalog()
                >>> attributes = catalog.get_bandset_catalog_attributes(
                bandset_number=1)
                >>> print(attributes)
                [(1, 'example', '2000-12-31', 'None', '20000101_1605495327_293')]
        """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        cfg.logger.log.debug(
            'number: %s; attribute: %s' % (bandset_number, attribute)
        )
        record = self.bandsets_table[
            self.bandsets_table['bandset_number'] == bandset_number]
        if len(record) == 0:
            return None
        else:
            if attribute is None:
                return record.tolist()
            else:
                result = record[attribute][0]
                cfg.logger.log.debug('result: %s' % result)
                return result

    def get_bandset(
            self, bandset_number: int, attribute: Optional[str] = None
    ) -> Union[None, str, list, BandSet]:
        """Get BandSet or BandSet attributes by number of the BandSet.

        This function gets the BandSet or BandSet attributes by the number
            of the BandSet.

        Args:
            bandset_number: number of the BandSet
            attribute: string of the attribute name, if None then the
                BandSet is returned

        Returns:
            The BandSet band attributes or the BandSet.

        Examples:
            Get the 'name' attribute of bands of the BandSet 1.
                >>> catalog = BandSetCatalog()
                >>> names = catalog.get_bandset(bandset_number=1,
                attribute='name')
                >>> print(names)
                ['file1', 'file2', 'file3']

            Get the BandSet 1.
                >>> catalog = BandSetCatalog()
                >>> bandset_1 = catalog.get_bandset(bandset_number=1)
                >>> print(bandset_1)
                BandSet
        """
        if bandset_number is None:
            bandset_number = self.current_bandset
        record = self.bandsets_table[
            self.bandsets_table['bandset_number'] == bandset_number]
        if len(record['uid']) == 0:
            cfg.logger.log.debug('get_bandset: None')
            return None
        uid = record['uid'][0]
        bs = self.bandsets[uid]
        if attribute is None:
            result = bs
        else:
            result = bs.get_band_attributes(attribute)
        cfg.logger.log.debug('bandset_number: %i' % bandset_number)
        return result

    def get_bandset_bands_by_attribute(
            self, bandset_number: int, attribute: str,
            attribute_value: Union[float, int, str],
            output_number: Optional[bool] = False
    ) -> Union[None, list, numpy.recarray]:
        """Get BandSet bands from the attributes thereof.

         This function gets BandSet bands or the band number of bands whose
            attributes match an attribute value.

         Args:
             bandset_number: number of the BandSet
             attribute: string of the attribute name
             attribute_value: value of the band attribute to find
             output_number: if True then the output is the band number list,
                if False then the output is the band array

         Returns:
             The band number list or the array of bands.

         Examples:
             Get the 'name' attribute of bands of the BandSet 1.
                 >>> catalog = BandSetCatalog()
                 >>> band_number = catalog.get_bandset_bands_by_attribute(bandset_number=1,attribute='wavelength',attribute_value=0.443,output_number=True)
                 >>> print(band_number)
                 [2]

             Get the BandSet 1.
                 >>> catalog = BandSetCatalog()
                 >>> band_x = catalog.get_bandset_bands_by_attribute(bandset_number=1,attribute='wavelength',attribute_value=0.443,output_number=True)
                 >>> print(band_x)
                 [(1, 1, '/data/file1.tif', '/data/file1.tif', 'file1', 0.443, 'µm (1 E-6m)', ...]
         """  # noqa: E501
        cfg.logger.log.debug('bandset_number: %s' % bandset_number)
        bs = self.get(bandset_number)
        result = bs.get_bands_by_attributes(
            attribute=attribute, attribute_value=attribute_value,
            output_as_number=output_number
        )
        return result

    def reset(self):
        """Resets the BandSets Catalog.

        This function resets the BandSets Catalog removing all BandSets and
            creating an empty one.

        Examples:
            Reset the BandSets catalog.
                >>> catalog = BandSetCatalog()
                >>> catalog.reset()
        """
        self.current_bandset = 1
        self.bandsets = {}
        self.bandsets_table = None
        # empty BandSet
        self._empty_bandset()
        cfg.logger.log.debug('reset')

    def create_bandset(
            self, paths: Union[list, str] = None,
            band_names: Optional[list] = None,
            wavelengths: Optional[list or str] = None,
            unit: Optional[str] = None,
            multiplicative_factors: Optional[list] = None,
            additive_factors: Optional[list] = None,
            date: Optional[list or str] = None,
            bandset_number: Optional[int] = None,
            insert: Optional[bool] = False,
            root_directory: Optional[str] = None,
            bandset_name: Optional[str] = None,
            box_coordinate_list: Optional[list] = None
    ) -> BandSet:
        """Creates a BandSet adding it to the catalog.

        This function creates a BandSet adding it to the catalog,
        by inserting or replacing the BandSet at specific BandSet number.
        Wavelength is defined by providing a list of values for each band,
        or a string of sensors names as defined in configurations
        satWlList such as:
         
        - Sentinel-2;
        - Landsat8;
        - Landsat5;
        - ASTER.

        Wavelength unit is defined by a string such as:
        
        - band number unitless band order as defined in the BandSet;
        - µm (1 E-6m) micrometers;
        - nm (1 E-9m) nanometers.

        Args:
            paths: list of file paths or a string of directory path; also, a list of directory path 
                and name filter is accepted.
            band_names: list of raster names used for identifying the bands, 
                if None then the names are automatically extracted from file names.
            wavelengths: list of center wavelengths of bands or string of sensor names (also partial).
            unit: wavelength unit as string
            multiplicative_factors: multiplicative factors for bands during calculations.
            additive_factors: additive factors for bands during calculations.
            date: list of date strings, or single date string (format YYYY-MM-DD) 
                or string defined in configurations date_auto to detect date from directory name.
            bandset_number: number of the BandSet, replacing an existing BandSet with the same number.
            insert: if True insert the BandSet at bandset_number, if False replace the BandSet number.
            root_directory: root directory for relative path.
            bandset_name: name of the BandSet.
            box_coordinate_list: list of coordinates [left, top, right, bottom] 
                to create a virtual subset.

        Returns:
            The created BandSet.

        Examples:
            Create a first BandSet from a file list with files inside a data directory, setting root_directory, defining the BandSet date, and the wavelenght from satellite name.
                >>> catalog = BandSetCatalog()
                >>> file_list = ['file1.tif', 'file2.tif', 'file3.tif']
                >>> bandset_date = '2021-01-01'
                >>> data_directory = 'data'
                >>> bandset = catalog.create_bandset(
                ... paths=file_list, wavelengths=['Sentinel-2'], date=bandset_date, 
                ... root_directory=data_directory
                ... )

            Create a new BandSet from a file list with files inside a data directory (setting root_directory), defining the BandSet date, and explicitly defining the BandSet number.
                >>> catalog = BandSetCatalog()
                >>> file_list = ['file1.tif', 'file2.tif', 'file3.tif']
                >>> bandset_date = '2021-01-01'
                >>> data_directory = 'data'
                >>> bandset = catalog.create_bandset(
                ... paths=file_list, wavelengths=['Sentinel-2'], date=bandset_date, 
                ... bandset_number=2, root_directory=data_directory
                ... )
                
           Passing a directory with file name filter and the wavelenght from satellite name.
                >>> bandset = catalog.create_bandset(['directory_path', 'tif'], wavelengths='Sentinel-2')
        """  # noqa: E501
        cfg.logger.log.debug('start')
        bst = BandSet.create(
            paths, band_names=band_names, wavelengths=wavelengths, unit=unit,
            multiplicative_factors=multiplicative_factors,
            additive_factors=additive_factors, dates=date,
            root_directory=root_directory, name=bandset_name,
            box_coordinate_list=box_coordinate_list
        )
        self.add_bandset(
            bandset=bst, bandset_number=bandset_number, insert=insert
        )
        cfg.logger.log.debug('end')
        return bst

    def _load(self, bandset_catalog, current_bandset: int = 1):
        """Loads a BandSet Catalog.

        This function loads a BandSet Catalog, and optionally defines the
        current BandSet.

        Args:
            bandset_catalog: BandSet Catalog
            current_bandset: optional integer number of current BandSet
        """
        cfg.logger.log.debug('start')
        self.reset()
        # copy BandSet Catalog
        bandset_catalog_copy = deepcopy(bandset_catalog)
        if bandset_catalog_copy.bandsets is not None:
            self.bandsets = bandset_catalog_copy.bandsets
        if bandset_catalog_copy.bandsets_table is None:
            # empty BandSet
            self._empty_bandset()
        else:
            self.bandsets_table = bandset_catalog_copy.bandsets_table
        self.current_bandset = current_bandset
        cfg.logger.log.debug('end')

    def move_band_in_bandset(
            self, band_number_input: int, band_number_output: int,
            bandset_number: Optional[int] = None, wavelength=True
    ):
        """Moves band in Bandset.

         This function reorders a band in a Bandset.

         Args:
            band_number_input: number of band to be moved.
            band_number_output: position of the moved band.
            bandset_number: number of BandSet; if None, current BandSet is used.
            wavelength: if True, keep the wavelength attributes of the band order.

         Examples:
             Move the second band of the BandSet 1 to position 5.
                 >>> catalog = BandSetCatalog()
                 >>> catalog.move_band_in_bandset(
                 ... bandset_number=1, band_number_input=2, band_number_output=5
                 ... )
         """  # noqa: E501
        cfg.logger.log.debug('start')
        if bandset_number is None:
            bandset_number = self.current_bandset
        bandset = self.get_bandset_by_number(bandset_number)
        wavelength_list = bandset.get_wavelengths()
        # temporary set -1 to moved band
        bandset.bands['band_number'][
            bandset.bands['band_number'] == band_number_input] = -1
        # move down bands between band_number_input and band_number_output
        if band_number_input < band_number_output:
            bandset.bands['band_number'][
                (bandset.bands['band_number'] > band_number_input) & (
                        bandset.bands['band_number'] <= band_number_output)] \
                = bandset.bands['band_number'][
                      (bandset.bands['band_number'] > band_number_input) & (
                              bandset.bands[
                                  'band_number'] <= band_number_output)] - 1
        # move up bands between band_number_input and band_number_output
        elif band_number_input > band_number_output:
            bandset.bands['band_number'][
                (bandset.bands['band_number'] >= band_number_input) & (
                        bandset.bands['band_number'] < band_number_output)] = \
                bandset.bands['band_number'][
                    (bandset.bands['band_number'] >= band_number_input) & (
                            bandset.bands[
                                'band_number'] < band_number_output)] + 1
        # set band_number_output to moved band
        bandset.bands['band_number'][
            bandset.bands['band_number'] == -1] = band_number_output
        bandset.bands.sort(order='band_number')
        if wavelength:
            bandset.bands['wavelength'] = wavelength_list
        cfg.logger.log.debug('end')

    def remove_band_in_bandset(
            self, band_number: int, bandset_number: Optional[int] = None
    ):
        """Removes band in Bandset.

         This function removes a band in Bandset identified by a number.
         Automatically reorders other band numbers.

         Args:
            band_number: number of band to be removed.
            bandset_number: number of BandSet; if None, current BandSet is used.

         Examples:
             Remove the second band of the BandSet.
                 >>> catalog = BandSetCatalog()
                 >>> catalog.remove_band_in_bandset(
                 ... bandset_number=1, band_number=2
                 ... )
         """  # noqa: E501
        cfg.logger.log.debug('start')
        if bandset_number is None:
            bandset_number = self.current_bandset
        bandset = self.get_bandset(bandset_number)
        # remove from table
        bandset.bands = bandset.bands[
            bandset.bands['band_number'] != band_number]
        # reorder other band numbers
        if band_number < bandset.get_band_count():
            bandset.bands['band_number'][
                bandset.bands['band_number'] > band_number] = \
                bandset.bands['band_number'][
                    bandset.bands['band_number'] > band_number] - 1
        cfg.logger.log.debug('end')

    def add_band_to_bandset(
            self, path: str, bandset_number: Optional[int] = None,
            band_number: Optional[int] = None,
            raster_band: Optional[int] = None, band_name: Optional[str] = None,
            date: Optional[str] = None, unit: Optional[str] = None,
            root_directory: Optional[str] = None,
            multiplicative_factor: Optional[int] = None,
            additive_factor: Optional[int] = None,
            wavelength: Optional[float] = None
    ):
        """Adds a new band to BandSet.

        This function creates a new band and adds it to a BandSet.

        Args:
            path: file path.
            band_name: raster name used for identifying the bands.
            wavelength: center wavelengths of band.
            unit: wavelength unit as string
            multiplicative_factor: multiplicative factor for bands during calculations.
            additive_factor: additive factors for band during calculations.
            date: date string (format YYYY-MM-DD).
            bandset_number: number of the BandSet; if None, the band is added to the current BandSet.
            root_directory: root directory for relative path.
            raster_band: raster band number.
            band_number: number of band in BandSet.

        Examples:
            Add a band to BandSet 1.
                >>> catalog = BandSetCatalog()
                >>> catalog.add_band_to_bandset(
                    path='file1.tif', bandset_number=1, band_number=1, raster_band=1
                ... )
        """  # noqa: E501
        cfg.logger.log.debug('start')
        if bandset_number is None:
            bandset_number = self.current_bandset
        self.get_bandset_by_number(bandset_number).add_new_band(
            path=path, band_number=band_number, raster_band=raster_band,
            band_name=band_name, date=date,
            root_directory=root_directory,
            multiplicative_factor=multiplicative_factor,
            additive_factor=additive_factor,
            wavelength=wavelength, unit=unit
        )
        cfg.logger.log.debug('end')

    def sort_bands_by_wavelength(self, bandset_number: Optional[int] = None):
        """Sorts bands by wavelength.

         This function numerically sorts bands in a BandSet by wavelength center.

         Args:
            bandset_number: number of BandSet; if None, current BandSet is used.

         Examples:
            Sort bands in BandSet 1.
                >>> catalog = BandSetCatalog()
                >>> catalog.sort_bands_by_wavelength(bandset_number=1)
         """  # noqa: E501
        cfg.logger.log.debug('bandset_number: %s' % bandset_number)
        if bandset_number is None:
            bandset_number = self.current_bandset
        self.get_bandset(bandset_number).sort_bands_by_wavelength()

    def add_bandset(
            self, bandset: BandSet, bandset_number: Optional[int] = None,
            insert=False
    ):
        """Adds a BandSet to Catalog.

         This function adds a previously created BandSet to BandSet Catalog.

         Args:
            bandset: the BandSet to be added.
            bandset_number: number of BandSet; if None, current BandSet is used.
            insert: if True insert the BandSet at bandset_number (other BandSets are moved), 
                if False replace the BandSet number.
                
         Examples:
            Insert a BandSet as BandSet 1.
                >>> catalog = BandSetCatalog()
                >>> catalog.add_bandset(bandset_number=1, insert=True
                ... )
         """  # noqa: E501
        cfg.logger.log.debug('bandset_number: %s' % bandset_number)
        # copy BandSet
        bandset_copy = deepcopy(bandset)
        bandset_count = self.get_bandset_count()
        if not insert:
            if bandset_number is None:
                bandset_number = self.current_bandset
            # new BandSet
            else:
                # add new BandSet if bandset_number > BandSet count
                if bandset_number > bandset_count:
                    bandset_number = bandset_count + 1
            if bandset_number <= bandset_count:
                # temporary set -1 to replaced bandset
                self.bandsets_table['bandset_number'][
                    (self.bandsets_table[
                         'bandset_number'] == bandset_number)] = -1
        # check existing BandSet
        else:
            try:
                if bandset_number is None:
                    bandset_number = self.current_bandset
                # reorder band sets
                if (bandset_number <= bandset_count
                        and self.bandsets_table is not None):
                    self.bandsets_table['bandset_number'][
                        self.bandsets_table[
                            'bandset_number'] >= bandset_number] = \
                        self.bandsets_table['bandset_number'][
                            self.bandsets_table[
                                'bandset_number'] >= bandset_number] + 1
                elif bandset_number >= bandset_count + 1:
                    bandset_number = bandset_count + 1
            except Exception as err:
                cfg.logger.log.error(str(err))
                messages.error(str(err))
        # add BandSet to dictionary by uid
        bandset_copy.uid = bandset_copy.generate_uid()
        self.bandsets[bandset_copy.uid] = bandset_copy
        if bandset_copy.box_coordinate_list is not None:
            box_coordinate_left = bandset_copy.box_coordinate_list[0]
            box_coordinate_top = bandset_copy.box_coordinate_list[1]
            box_coordinate_right = bandset_copy.box_coordinate_list[2]
            box_coordinate_bottom = bandset_copy.box_coordinate_list[3]
        else:
            box_coordinate_left = box_coordinate_top = box_coordinate_right \
                = box_coordinate_bottom = None
        self.bandsets_table = tm.create_bandset_catalog_table(
            bandset_number=bandset_number,
            root_directory=bandset_copy.root_directory, date=bandset_copy.date,
            bandset_uid=bandset_copy.uid, bandset_name=bandset_copy.name,
            previous_catalog=self.bandsets_table,
            crs=bandset_copy.crs, box_coordinate_left=box_coordinate_left,
            box_coordinate_top=box_coordinate_top,
            box_coordinate_right=box_coordinate_right,
            box_coordinate_bottom=box_coordinate_bottom
        )
        if not insert and bandset_number <= bandset_count:
            self._remove_bandset(-1, reorder=False)

    def move_bandset(
            self, bandset_number_input: int, bandset_number_output: int
    ):
        """Moves a BandSet.

         This function reorders a BandSet in the BandSet Catalog.

         Args:
            bandset_number_input: the BandSet number to be moved.
            bandset_number_output: the new poistion of BandSet.

         Examples:
            Move BandSet 1 to position 3.
                >>> catalog = BandSetCatalog()
                >>> catalog.move_bandset(bandset_number_input=1, bandset_number_output=3)
         """  # noqa: E501
        cfg.logger.log.debug(
            'bandset_number_input: %s' % str(bandset_number_input)
        )
        # temporary set -1 to moved bandset
        self.bandsets_table['bandset_number'][
            (self.bandsets_table[
                 'bandset_number'] == bandset_number_input)] = -1
        # move down BandSets between bandset_number_input and
        # bandset_number_output
        if bandset_number_input < bandset_number_output:
            self.bandsets_table['bandset_number'][
                (self.bandsets_table['bandset_number'] >
                 bandset_number_input) & (
                        self.bandsets_table['bandset_number'] <=
                        bandset_number_output)] = \
                self.bandsets_table['bandset_number'][
                    (self.bandsets_table['bandset_number'] >
                     bandset_number_input) & (
                            self.bandsets_table['bandset_number'] <=
                            bandset_number_output)] - 1
        # move up BandSets between bandset_number_input and
        # bandset_number_output
        elif bandset_number_input > bandset_number_output:
            self.bandsets_table['bandset_number'][
                (self.bandsets_table['bandset_number'] >=
                 bandset_number_output) & (
                        self.bandsets_table['bandset_number']
                        < bandset_number_input)] = \
                self.bandsets_table[
                    'bandset_number'][
                    (self.bandsets_table['bandset_number'] >=
                     bandset_number_output) & (
                            self.bandsets_table['bandset_number'] <
                            bandset_number_input)] + 1
        cfg.logger.log.info(
            'bandset bandset_number_input: %s; bandset '
            'bandset_number_output: %s'
            % (str(bandset_number_input), str(bandset_number_output))
        )
        # set bandset_number to moved BandSet
        self.bandsets_table['bandset_number'][
            (self.bandsets_table['bandset_number'] == -1)] = \
            bandset_number_output
        # sort
        self.bandsets_table.sort(order='bandset_number')

    def remove_bandset(self, bandset_number: int):
        """Removes a BandSet.

         This function removes a BandSet by the number thereof.

         Args:
            bandset_number: the BandSet number to be removed.

         Examples:
            Remove BandSet 2.
                >>> catalog = BandSetCatalog()
                >>> catalog.remove_bandset(2)
         """  # noqa: E501
        self._remove_bandset(bandset_number, reorder=True)

    def _remove_bandset(self, bandset_number: int, reorder=True):
        """Function to remove a BandSet by number and optional reorder."""
        cfg.logger.log.debug('bandset_number: %s' % str(bandset_number))
        if bandset_number > self.get_bandset_count():
            cfg.logger.log.warning('bandset_number: %s' % str(bandset_number))
            messages.warning('bandset_number: %s' % str(bandset_number))
        else:
            record = self.bandsets_table[
                self.bandsets_table['bandset_number'] == bandset_number]
            uid = record['uid'][0]
            # remove from table
            self.bandsets_table = self.bandsets_table[
                self.bandsets_table['uid'] != uid]
            # move down BandSets above removed BandSet
            if reorder and bandset_number < self.get_bandset_count():
                self.bandsets_table['bandset_number'][
                    (self.bandsets_table['bandset_number'] >
                     bandset_number)] = \
                    self.bandsets_table[
                        'bandset_number'][(self.bandsets_table[
                                               'bandset_number'] >
                                           bandset_number)] - 1
            # remove from dictionary
            self.bandsets.pop(str(uid), None)
            if self.get_bandset_count() == 0:
                # empty BandSet
                self._empty_bandset()

    def _empty_bandset(self):
        """Function to create an empty BandSet."""
        bandset = BandSet.create()
        bandset_number = 1
        # add BandSet to dictionary by uid
        bandset.uid = bandset.generate_uid()
        self.bandsets[bandset.uid] = bandset
        self.bandsets_table = tm.create_bandset_catalog_table(
            bandset_number=bandset_number,
            root_directory=bandset.root_directory, date=bandset.date,
            bandset_uid=bandset.uid, bandset_name=bandset.name,
            previous_catalog=self.bandsets_table
        )
        cfg.logger.log.debug('empty bandset')

    @staticmethod
    def get_band_list(
            bandset: Union[int, list, BandSet] = None,
            bandset_catalog: Optional = None
            ) -> list:
        """Gets band list.

        This function gets band list from several types of input.

        Args:
            bandset: input of type BandSet or list of paths or integer number of BandSet.
            bandset_catalog: optional type BandSetCatalog if bandset argument is a number.
            
        Returns:
            If argument bandset is a BandSet or a BandSet number, 
                returns list of bands in the BandSet; if argument bandset is 
                already a list, returns the same list.
            
        Examples:
            Get band list of BandSet 1.
                >>> # import Remotior Sensus and start the session
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()
                >>> catalog = BandSetCatalog()
                >>> list_1 = rs.bandset_catalog.get_band_list(1, catalog)
                >>> # which is equivalent to
                >>> bandset_1 = bandset_catalog.get_bandset(1)
                >>> list_2 = bandset_1.get_absolute_paths()
        """  # noqa: E501
        # BandSet
        if type(bandset) is BandSet:
            band_list = bandset.get_absolute_paths()
        # list of paths
        elif type(bandset) is list:
            band_list = bandset
        # number of BandSet
        elif bandset_catalog is not None:
            b = bandset_catalog.get_bandset(bandset)
            band_list = b.get_absolute_paths()
        else:
            band_list = None
        cfg.logger.log.debug('band list: %s' % str(band_list))
        return band_list

    def iterate_bandset_bands(self, attribute: str) -> list:
        """Iterates BandSet attributes.

        This function gets BandSet attributes iterating all the BandSets 
            in the Catalog.

        Args:
            attribute: attribute name.

        Returns:
            List of attributes

        Examples:
            Get BandSet names.
                >>> catalog = BandSetCatalog()
                >>> names = catalog.iterate_bandset_bands('name')
        """  # noqa: E501
        attributes = []
        for i in range(1, self.get_bandset_count() + 1):
            attributes.extend(self.get_bandset(i, attribute))
        return attributes

    def find_bandset_names_in_list(
            self, names: list, lower=True, output_number=True,
            exact_match=False
    ) -> list:
        """Finds BandSet names.

        This function finds BandSet names in a list and return BandSets 
            or BandSet numbers.

        Args:
            names: list of string names.
            lower: if True, finds by lowering all the names.
            output_number: if True returns the number of the band, if False returns the BandSet.
            exact_match: if True, names are compared by equality.

        Returns:
            if output_number is True, returns list of band number;
            if output_number is False, returns list of BandSets.

        Examples:
            Find BandSet from name list.
                >>> catalog = BandSetCatalog()
                >>> name_list = ['name1', 'name2']
                >>> bandsets = catalog.find_bandset_names_in_list(names=name_list)
        """  # noqa: E501
        bandset_names = self.bandsets_table['bandset_name']
        bandset_list = []
        # iterate names
        for bandset_name in bandset_names.tolist():
            bandset_name_check = bandset_name
            for name in names:
                if lower:
                    name = name.lower()
                    bandset_name_check = bandset_name.lower()
                if exact_match:
                    if name == bandset_name_check:
                        bandset_list.append(
                            self.get_bandset_by_name(
                                bandset_name, output_number=output_number
                            )
                        )
                        break
                else:
                    if name in bandset_name_check:
                        bandset_list.append(
                            self.get_bandset_by_name(
                                bandset_name, output_number=output_number
                            )
                        )
                        break
        return bandset_list

    def set_root_directory(
            self, root_directory: str, bandset_number: Optional[int] = None
    ):
        """Sets BandSet root directory.

        Sets BandSet root directory for relative path.

        Args:
            root_directory: root directory path.
            bandset_number: number of BandSet; if None, current BandSet is used.

        Examples:
            Set BandSet 1 root directory.
                >>> catalog = BandSetCatalog()
                >>> catalog.set_root_directory(bandset_number=1, root_directory='data_directory')
        """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        self.get_bandset(bandset_number).root_directory = root_directory
        self.bandsets_table['root_directory'][self.bandsets_table[
                                                  'bandset_number'] ==
                                              bandset_number] = root_directory

    def set_date(self, date: str, bandset_number: Optional[int] = None):
        """Sets BandSet date.

        Sets BandSet date.

        Args:
            date: date string (format YYYY-MM-DD).
            bandset_number: number of BandSet; if None, current BandSet is used.

        Examples:
            Set BandSet 1 date.
                >>> catalog = BandSetCatalog()
                >>> catalog.set_date(bandset_number=1, date='2021-01-01')
        """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        self.get_bandset(bandset_number).date = date
        self.bandsets_table['date'][
            self.bandsets_table['bandset_number'] == bandset_number] = date

    def set_name(self, name: str, bandset_number: int = None):
        """Sets BandSet name.

        Sets BandSet name.

        Args:
            name: name.
            bandset_number: number of BandSet; if None, current BandSet is used.

        Examples:
            Set BandSet 1 name.
                >>> catalog = BandSetCatalog()
                >>> catalog.set_name(bandset_number=1, name='example')
        """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        self.get_bandset(bandset_number).name = name
        self.bandsets_table['bandset_name'][
            self.bandsets_table['bandset_number'] == bandset_number] = name

    def set_box_coordinate_list(
            self, box_coordinate_list: list,
            bandset_number: Optional[int] = None
    ):
        """Sets BandSet box coordinate list.

        Sets BandSet box coordinate list  for virtual subset.

        Args:
            box_coordinate_list: list of coordinates [left, top, right, bottom] to create a virtual subset.
            bandset_number: number of BandSet; if None, current BandSet is used.

        Examples:
            Set BandSet 1 coordinate list.
                >>> catalog = BandSetCatalog()
                >>> catalog.set_box_coordinate_list(
                ... bandset_number=1, box_coordinate_list=[230000, 4680000, 232000, 4670000]
                ... )
        """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        self.get_bandset(
            bandset_number
        ).box_coordinate_list = box_coordinate_list
        self.bandsets_table['box_coordinate_left'][
            self.bandsets_table['bandset_number'] == bandset_number] = \
            box_coordinate_list[0]
        self.bandsets_table['box_coordinate_top'][
            self.bandsets_table['bandset_number'] == bandset_number] = \
            box_coordinate_list[1]
        self.bandsets_table['box_coordinate_right'][
            self.bandsets_table['bandset_number'] == bandset_number] = \
            box_coordinate_list[2]
        self.bandsets_table['box_coordinate_bottom'][
            self.bandsets_table['bandset_number'] == bandset_number] = \
            box_coordinate_list[3]

    def get_root_directory(self, bandset_number: Optional[int] = None) -> str:
        """Gets BandSet root directory.

        Gets BandSet root directory.

        Args:
            bandset_number: number of BandSet; if None, current BandSet is used.

        Returns:
            Root directory string.
            
        Examples:
            Get BandSet 1 root directory.
                >>> catalog = BandSetCatalog()
                >>> catalog.get_root_directory(1)
        """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        return self.get_bandset_catalog_attributes(
            bandset_number=bandset_number, attribute='root_directory'
        )

    def get_date(self, bandset_number: Optional[int] = None) -> str:
        """Gets BandSet date.

        Gets BandSet date.

        Args:
            bandset_number: number of BandSet; if None, current BandSet is used.

        Returns:
            Date string.
            
        Examples:
            Get BandSet 1 date.
                >>> catalog = BandSetCatalog()
                >>> catalog.get_date(1)
        """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        return str(
            self.get_bandset_catalog_attributes(
                bandset_number=bandset_number, attribute='date'
            )
        )

    def get_name(self, bandset_number: Optional[int] = None) -> str:
        """Gets BandSet name.

        Gets BandSet name.

        Args:
            bandset_number: number of BandSet; if None, current BandSet is used.

        Returns:
            BandSet name.
            
        Examples:
            Get BandSet 1 date.
                >>> catalog = BandSetCatalog()
                >>> catalog.get_name(1)
        """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        return str(
            self.get_bandset_catalog_attributes(
                bandset_number=bandset_number, attribute='bandset_name'
            )
        )

    def get_band_count(
            self, bandset_number: Optional[int] = None
    ) -> Union[None, int]:
        """Gets band count.

        Gets band count of a BandSet.

        Args:
            bandset_number: number of BandSet; if None, current BandSet is used.

        Returns:
            count of bands.
            
        Examples:
            Get BandSet 1 band count.
                >>> catalog = BandSetCatalog()
                >>> catalog.get_band_count(1)
         """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        bandset = self.get(bandset_number=bandset_number)
        if bandset is None:
            return None
        else:
            count = bandset.get_band_count()
            cfg.logger.log.debug('band_count: %s' % str(count))
            return count

    def get_box_coordinate_list(
            self, bandset_number: int = None
    ) -> Union[None, list]:
        """Gets BandSet box coordinate list.

        Gets BandSet box coordinate list.

        Args:
            bandset_number: number of BandSet; if None, current BandSet is used.

        Returns:
            List of box coordinates [left, top, right, bottom].
            
        Examples:
            Get BandSet 1 band count.
                >>> catalog = BandSetCatalog()
                >>> catalog.get_box_coordinate_list(1)
        """  # noqa: E501
        if bandset_number is None:
            bandset_number = self.current_bandset
        bandset = self.get(bandset_number=bandset_number)
        if bandset is None:
            return None
        else:
            box_coordinate_list = [
                self.bandsets_table['box_coordinate_left'][
                    self.bandsets_table['bandset_number'] == bandset_number][
                    0],
                self.bandsets_table['box_coordinate_top'][
                    self.bandsets_table['bandset_number'] == bandset_number][
                    0],
                self.bandsets_table['box_coordinate_right'][
                    self.bandsets_table['bandset_number'] == bandset_number][
                    0],
                self.bandsets_table['box_coordinate_bottom'][
                    self.bandsets_table['bandset_number'] == bandset_number][
                    0]]
            return box_coordinate_list

    def create_band_string_list(self, bandset_number: int = None) -> list:
        """Creates band string list.

        Creates band string list such as "bandset1b1" for all the bands in a BandSet. 
        Used in tool :func:`~remotior_sensus.tools.band_calc`.

        Args:
            bandset_number: number of BandSet; if None, current BandSet is used.
            
        Returns:
            List of band string such as ["bandset1b1", "bandset1b2"].
            
        Examples:
            Get BandSet 1 band string list.
                >>> catalog = BandSetCatalog()
                >>> catalog.create_band_string_list(1)
         """  # noqa: E501
        number = self.get(bandset_number).bands['band_number']
        string_1 = numpy.char.add(
            '%s%s%s%s' % (cfg.variable_band_quotes, cfg.variable_bandset_name,
                          str(bandset_number), cfg.variable_band_name),
            number.astype('<U16')
        )
        string_2 = numpy.char.add(string_1, cfg.variable_band_quotes)
        return string_2.tolist()



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Tools to download files
"""

import datetime
import os
import urllib.request
from http.cookiejar import CookieJar

from remotior_sensus.core import configurations as cfg, messages
from remotior_sensus.util import read_write_files, files_directories


# get proxy handler
def get_proxy_handler(
        proxy_host, proxy_port, proxy_user=None, proxy_password=None
):
    if len(proxy_user) > 0:
        proxy_handler = urllib.request.ProxyHandler(
            {
                'http': 'http://{}:{}@{}:{}'.format(
                    proxy_user, proxy_password, proxy_host, proxy_port
                )
            }
        )
    else:
        proxy_handler = urllib.request.ProxyHandler(
            {'http': 'http://{}:{}'.format(proxy_host, proxy_port)}
        )
    return proxy_handler


# opener
def general_opener(
        proxy_host=None, proxy_port=None, proxy_user=None, proxy_password=None
):
    cookie_jar = CookieJar()
    if proxy_host:
        proxy_handler = get_proxy_handler(
            proxy_host, proxy_port, proxy_user,
            proxy_password
        )
        opener = urllib.request.build_opener(
            proxy_handler, urllib.request.HTTPCookieProcessor(cookie_jar)
        )
    else:
        opener = urllib.request.build_opener(
            urllib.request.HTTPCookieProcessor(cookie_jar)
        )
    return opener, cookie_jar


# open general url
def open_general_url(
        url, proxy_host=None, proxy_port=None, proxy_user=None,
        proxy_password=None
):
    opener, cookie_jar = general_opener(
        proxy_host=proxy_host, proxy_port=proxy_port, proxy_user=proxy_user,
        proxy_password=proxy_password
    )
    request = urllib.request.Request(url)
    try:
        response = opener.open(request)
    except Exception as err:
        response = None
        cfg.logger.log.error('{}; url:{}'.format(err, url))
        messages.error('{}; url:{}'.format(err, url))
    return response


# download file
def download_file(
        url, output_path, authentication_uri=None, user=None, password=None,
        proxy_host=None, proxy_port=None, proxy_user=None, proxy_password=None,
        progress=True, message=None, min_progress=0, max_progress=100
):
    cfg.logger.log.debug('url: %s' % url)
    if authentication_uri is None:
        auth_handler = None
    else:
        pass_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        pass_mgr.add_password(None, authentication_uri, user, password)
        auth_handler = urllib.request.HTTPBasicAuthHandler(pass_mgr)
    cookie_proc = urllib.request.HTTPCookieProcessor()
    # with proxy
    if proxy_host:
        proxy_handler = get_proxy_handler(
            proxy_host, proxy_port, proxy_user, proxy_password
        )
        if auth_handler:
            opener = urllib.request.build_opener(
                proxy_handler, cookie_proc, auth_handler
            )
        else:
            opener = urllib.request.build_opener(proxy_handler, cookie_proc)
    # without proxy
    elif auth_handler:
        opener = urllib.request.build_opener(cookie_proc, auth_handler)
    else:
        opener = urllib.request.build_opener(cookie_proc)
    urllib.request.install_opener(opener)
    try:
        url_request = urllib.request.urlopen(url)
        try:
            file_size = int(url_request.headers['Content-Length'])
            # size megabyte
            total_size = round(file_size / 1048576, 2)
        except Exception as err:
            str(err)
            file_size = 1
            total_size = 1
        block_size = 1024 * 1024
        # set initial speed for adaptive block size
        speed = 0
        adaptive_block_size = True
        # small files
        if block_size >= file_size:
            response = url_request.read()
            read_write_files.write_file(response, output_path, mode='wb')
        else:
            files_directories.create_parent_directory(output_path)
            with open(output_path, 'wb') as file:
                while True:
                    if adaptive_block_size:
                        start_time = datetime.datetime.now()
                    block_read = url_request.read(block_size)
                    if not block_read:
                        break
                    # adapt block size
                    if adaptive_block_size:
                        end_time = datetime.datetime.now()
                        time_delta = end_time - start_time
                        new_speed = block_size / time_delta.microseconds
                        if new_speed >= speed:
                            block_size = block_size + 1024 * 1024
                            start_time = end_time
                            speed = new_speed
                    if progress:
                        if message is None:
                            message = '({}/{} MB) {}'.format(
                                downloaded_part_size, total_size, url
                            )
                        downloaded_part_size = round(
                            int(os.stat(output_path).st_size) / 1048576, 2
                        )
                        step = int(
                            (max_progress - min_progress) *
                            downloaded_part_size / total_size + min_progress
                        )
                        percentage = int(
                            100 * downloaded_part_size / total_size
                        )
                        cfg.progress.update(
                            message=message, step=step, percentage=percentage
                        )
                    # write file
                    file.write(block_read)
        return True, output_path
    except Exception as err:
        cfg.logger.log.error('%s; url: %s' % (err, url))
        messages.error('%s; url: %s' % (err, url))
        return False, str(err)



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Session manager.

Core class that manages the Remotior Sensus' session. It defines fundamental
parameters such as available RAM and number of parallel processes.
Creates a temporary directory to store temporary files.
Exposes core functions and tools.
It includes a method to close the session removing temporary files
and stopping parallel processes.

Typical usage example:

    >>> # start the session
    >>> rs = Session()
    >>> # optionally set session parameters
    >>> rs.set(n_processes=2)
    >>> # close the session when processing is done
    >>> rs.close()
"""

import logging
from types import FunctionType
from typing import Optional

from remotior_sensus.core import configurations, messages, table_manager
from remotior_sensus.core.bandset import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.core.log import Log
from remotior_sensus.core.multiprocess_manager import Multiprocess
from remotior_sensus.core.progress import Progress
from remotior_sensus.core.spectral_signatures import SpectralSignaturesCatalog
from remotior_sensus.core.temporary import Temporary
from remotior_sensus.tools import (
    band_calc, band_classification, band_combination, band_dilation,
    band_erosion, band_neighbor_pixels, band_pca, band_sieve,
    cross_classification, download_products, band_mosaic, preprocess_products,
    raster_reclassification, raster_report, raster_to_vector
)
from remotior_sensus.util import dates_times, system_tools, files_directories


class Session(object):
    """Manages system parameters.

    This module allows for managing Remotior Sensus' session,
    setting fundamental processing parameters and exposing core functions 
    and tools.

    Attributes:
        configurations: module containing shared variables and functions
        bandset: access :func:`~remotior_sensus.core.bandset.BandSet` class
        bandset_catalog: access :func:`~remotior_sensus.core.bandset_catalog.BandSetCatalog` class
        spectral_signatures_catalog: access :func:`~remotior_sensus.core.spectral_signatures.SpectralSignaturesCatalog` class 
        output_manager: access :func:`~remotior_sensus.core.output_manager.OutputManager` class 
        table_manager: access functions of :func:`~remotior_sensus.core.table_manager` module
        dates_times: access dates and times utilities
        band_calc: tool :func:`~remotior_sensus.tools.band_calc`
        band_classification: tool :func:`~remotior_sensus.tools.band_classification`
        classifier: tool :func:`~remotior_sensus.tools.band_classification.Classifier`
        band_combination: tool :func:`~remotior_sensus.tools.band_combination`
        band_dilation: tool :func:`~remotior_sensus.tools.band_dilation`
        band_erosion: tool :func:`~remotior_sensus.tools.band_erosion`
        band_neighbor_pixels: tool :func:`~remotior_sensus.tools.band_neighbor_pixels`
        band_pca: tool :func:`~remotior_sensus.tools.band_pca`
        band_sieve: tool :func:`~remotior_sensus.tools.band_sieve`
        cross_classification: tool :func:`~remotior_sensus.tools.cross_classification`
        download_products: tool :func:`~remotior_sensus.tools.download_products`
        band_mosaic: tool :func:`~remotior_sensus.tools.band_mosaic`
        preprocess_products: tool :func:`~remotior_sensus.tools.preprocess_products`
        raster_reclassification: tool :func:`~remotior_sensus.tools.preprocess_products`
        raster_report: tool :func:`~remotior_sensus.tools.raster_report`
        raster_to_vector: tool :func:`~remotior_sensus.tools.raster_to_vector`

    Examples:
        Start a session
            >>> import remotior_sensus
            >>> rs = remotior_sensus.Session()

        Start a session defining number of parallel processes. and available RAM
            >>> import remotior_sensus
            >>> rs = remotior_sensus.Session(n_processes=4, available_ram=4096)

        Create a :func:`~remotior_sensus.core.bandset_catalog.BandSetCatalog`
            >>> catalog = rs.bandset_catalog()

        Run the tool for raster report
            >>> output = rs.raster_report(raster_path='file.tif', output_path='output.txt')

        Stop a session at the end to clear temporary directory
            >>> rs.close()
    """  # noqa: E501

    def __init__(
            self, n_processes: Optional[int] = 2, available_ram: int = 2048,
            temporary_directory: str = None,
            directory_prefix: str = None, log_level: int = 20,
            log_time: bool = True, progress_callback=None
    ):
        """Starts a session.

        Starts a new session setting fundamental parameters for processing.
        It sets the number of parallel processes (default 2) and available RAM
        (default 2048MB) to be used  in calculations.
        It starts the class Temporary to manage temporary files by creating a
        temporary directory with an optional name prefix.
        It starts the class Log for logging (with a default level INFO) and
        creates a logging formatter with the option to hide time.
        It starts the class Progress for displaying progress with a default
        callback function.
        A custom progress callback function can be passed optionally.

        The sessions also allows for accessing to the core functions
        and tools.

        In the end, the close() function should be called to clear
        the temporary directory and stop the parallel processes.

        Args:
            n_processes: number of parallel processes.
            available_ram: number of megabytes of RAM available to processes.
            temporary_directory: path to a temporary directory.
            directory_prefix: prefix of the name of the temporary directory.
            log_level: level of logging (10 for DEBUG, 20 for INFO).
            log_time: if True, logging includes the time.
            progress_callback: function for progress callback.

        Examples:
            Start a session
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()
        """
        configurations.n_processes = n_processes
        configurations.available_ram = available_ram
        # create temporary directory
        temp = Temporary()
        if directory_prefix is None:
            directory_prefix = configurations.root_name
        configurations.temp = temp.create_root_temporary_directory(
            prefix=directory_prefix, directory=temporary_directory
        )
        # create logger
        if log_level is None:
            log_level = logging.INFO
        self.log_level = log_level
        configurations.logger = Log(
            directory=configurations.temp.dir, level=self.log_level,
            time=log_time
        )
        # start progress
        progress = Progress()
        if progress_callback is None:
            progress_callback = Progress.print_progress_replace
        configurations.progress = progress.start(callback=progress_callback)
        system_tools.get_system_info()
        check = _check_dependencies(configurations)
        if check:
            self.configurations = configurations
            # create multiprocess instance
            self.configurations.multiprocess = Multiprocess(n_processes)
            # available core tools
            self.bandset = BandSet
            self.bandset_catalog = BandSetCatalog
            self.spectral_signatures_catalog = SpectralSignaturesCatalog
            self.output_manager = OutputManager
            self.table_manager = table_manager
            # available tools
            self.band_calc = band_calc.band_calc
            self.band_classification = band_classification.band_classification
            self.classifier = band_classification.Classifier
            self.band_combination = band_combination.band_combination
            self.band_dilation = band_dilation.band_dilation
            self.band_erosion = band_erosion.band_erosion
            self.band_neighbor_pixels = \
                band_neighbor_pixels.band_neighbor_pixels
            self.band_pca = band_pca.band_pca
            self.band_sieve = band_sieve.band_sieve
            self.cross_classification = \
                cross_classification.cross_classification
            self.download_products = download_products
            self.band_mosaic = band_mosaic.band_mosaic
            self.preprocess_products = preprocess_products
            self.raster_reclassification = \
                raster_reclassification.raster_reclassification
            self.raster_report = raster_report.raster_report
            self.raster_to_vector = raster_to_vector.raster_to_vector
            self.dates_times = dates_times
        else:
            self.configurations = None

    def close(self, log_path: str = None):
        """Closes a Session.

        This function closes current session by deleting the temporary files
        and stopping parallel processes.

        Args:
            log_path: path where the log file is saved

        Examples:
            Given that a session was previously started
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()

            Set the number of parallel processes and available RAM
                >>> rs.set(n_processes=8, available_ram=20480)

            Set the logging level to DEBUG
                >>> rs.set(log_level=10)
        """
        if log_path:
            try:
                files_directories.copy_file(
                    self.configurations.logger.file_path, log_path
                )
            except Exception as err:
                str(err)
        self.configurations.temp.clear()
        self.configurations.multiprocess.stop()

    def set(
            self, n_processes: int = None, available_ram: int = None,
            temporary_directory: str = None,
            directory_prefix: str = None, log_level: int = None,
            log_time: bool = None, progress_callback: FunctionType = None
    ):
        """Sets or changes the parameters of an existing Session.

        Sets the parameters of an existing Session such as number
        of processes or temporary directory.

        Args:
            n_processes: number of parallel processes.
            available_ram: number of megabytes of RAM available to processes.
            temporary_directory: path to a temporary directory.
            directory_prefix: prefix of the name of the temporary directory.
            log_level: level of logging (10 for DEBUG, 20 for INFO).
            log_time: if True, logging includes the time.
            progress_callback: function for progress callback.

        Examples:
            Given that a session was previously started
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()

            Stop a session
                >>> rs.close()

            Stop a session saving also the log to a file
                >>> rs.close(log_path='file.txt')
        """
        if n_processes:
            self.configurations.n_processes = n_processes
            check = _check_dependencies(self.configurations)
            if check:
                self.configurations.multiprocess.stop()
                self.configurations.multiprocess = Multiprocess(n_processes)
            else:
                self.configurations = None
                return
        if available_ram:
            self.configurations.available_ram = available_ram
        if temporary_directory:
            self.configurations.temp.clear()
            # create temporary directory
            temp = Temporary()
            if directory_prefix is None:
                directory_prefix = self.configurations.root_name
            self.configurations.temp = temp.create_root_temporary_directory(
                prefix=directory_prefix, directory=temporary_directory
            )
        if log_level:
            self.log_level = log_level
            if log_time is None:
                log_time = True
            # create logger
            self.configurations.logger = Log(
                directory=self.configurations.temp.dir, level=self.log_level,
                time=log_time
            )
        elif log_time:
            # create logger
            self.configurations.logger = Log(
                directory=self.configurations.temp.dir, level=self.log_level,
                time=log_time
            )
        if progress_callback:
            # start progress
            progress = Progress()
            self.configurations.progress = progress.start(
                callback=progress_callback
            )
        self.configurations.logger.log.info(
            'n_processes: %s; ram: %s; temp.dir: %s'
            % (self.configurations.n_processes,
               self.configurations.available_ram,
               self.configurations.temp.dir)
        )


def _check_dependencies(configuration_module: configurations) -> bool:
    """Checks the dependencies.

    Checks the dependencies and returns a boolean.

    Args:
        configuration_module: module configurations used for logging.
    """
    check = True
    try:
        import os
        try:
            import numpy
        except Exception as err:
            configuration_module.logger.log.error(str(err))
            messages.error('dependency error: numpy')
            check = False
        try:
            from scipy import signal
        except Exception as err:
            configuration_module.logger.log.error(str(err))
            messages.error('dependency error: scipy')
            check = False
        try:
            import torch
        except Exception as err:
            configuration_module.logger.log.error(str(err))
            messages.error('dependency error: pytorch')
        if configuration_module.gdal_path is not None:
            os.add_dll_directory(configuration_module.gdal_path)
        try:
            from osgeo import gdal
        except Exception as err:
            configuration_module.logger.log.error(str(err))
            messages.error('dependency error: gdal')
            check = False
    except Exception as err:
        configuration_module.logger.log.error(str(err))
        messages.error(str(err))
        check = False
    return check
